            (* +--------------------------------------+
               |  Caclul de polynome caractristique  |
               +--------------------------------------+ *)

#open "num";;
let zero = num_of_int(0);;
let un   = num_of_int(1);;

                     (* +---------------------+
                        |  Mthode de Fadeev  |
                        +---------------------+ *)

let tr(m) =
  let n = vect_length(m) in
  let t = ref(zero)      in
  for i=0 to n-1 do t := !t +/ m.(i).(i) done;
  !t
;;

let prod a b =
  let n = vect_length(a)       in
  let p = vect_length(a.(0))   in
  let q = vect_length(b.(0))   in
  let c = make_matrix n q zero in
  for i=0 to n-1 do for j=0 to q-1 do
    for k=0 to p-1 do c.(i).(j) <- c.(i).(j) +/ a.(i).(k)*/b.(k).(j) done
  done done;
  c
;;

let add_diag m x =
  let n = vect_length(m)       in
  let p = make_matrix n n zero in
  for i=0 to n-1 do for j=0 to n-1 do p.(i).(j) <- m.(i).(j) done done;
  for i=0 to n-1 do p.(i).(i) <- p.(i).(i) +/ x done;
  p
;;

let poca_1 m =
  let n = vect_length(m)       in
  let p = make_vect (n+1) zero in
  p.(n) <- un;
  let a = ref(m) in
  for i=1 to n do
    p.(n-i) <- minus_num ((tr !a) // (num_of_int i));
    a := prod m (add_diag !a p.(n-i))
  done;
  p
;;
  

let m = map_vect (map_vect num_of_int)
	  [| [| 1;2;3 |];
             [| 4;5;6 |];
             [| 7;8;9 |] |];;
poca_1(m);;


                    (* +-----------------------+
                       |  Mthode de Lagrange  |
                       +-----------------------+ *)
 
let copie m =
  let n = vect_length(m)       in
  let p = vect_length(m.(0))   in
  let t = make_matrix n p zero in
  for i=0 to n-1 do for j=0 to p-1 do t.(i).(j) <- m.(i).(j) done done;
  t
;;

(* Dterminant par pivot *)
let det(t) =
  let m = copie(t)          in
  let n = vect_length(m)    in
  let d = ref(un)           in
  let j = ref(0)            in
  while (!j < n) && (!d <>/ zero) do
    let i = ref(!j) in
    while(!i < n) && (m.(!i).(!j) =/ zero) do i := !i+1 done;
    if !i < n then begin
      if !i > !j then begin
        for k = !j to n-1 do
	  let x = m.(!i).(k) in m.(!i).(k) <- m.(!j).(k); m.(!j).(k) <- x
	done;
        d := minus_num !d;
      end;
      d := !d */ m.(!j).(!j);
      for k = n-1 downto !j do m.(!j).(k) <- m.(!j).(k)//m.(!j).(!j) done;
      for i = !j+1 to n-1 do
        for k = n-1 downto !j do m.(i).(k) <- m.(i).(k) -/ m.(i).(!j)*/m.(!j).(k) done
      done
    end
    else d := zero;   
    j := !j + 1;
  done;
  !d
;;

det(add_diag m (num_of_int 3));;

(* Evaluation d'un polynme *)
let horner p x =
  let n = vect_length(p) in
  let y = ref(zero)      in
  for i=n-1 downto 0 do y := !y */ x +/ p.(i) done;
  !y
;;
  

(* Polynme d'interpolation de Lagrange *)
let lagrange x y =
  let n = vect_length(x)   in
  let p = make_vect n zero in
  let q = make_vect n zero in

  p.(0) <- y.(0); q.(0) <- un;
  for i=1 to n-1 do
    for j = i downto 1 do q.(j) <- q.(j-1) -/ x.(i-1)*/q.(j) done;
    q.(0) <- minus_num(x.(i-1) */ q.(0));
    let z = (y.(i) -/ (horner p x.(i))) // (horner q x.(i)) in
    for j=0 to i do p.(j) <- p.(j) +/ z*/q.(j) done;
  done;
  p
;;

lagrange (map_vect num_of_int [|1;2;3;4|])
         (map_vect num_of_int [|1;-3;2;0|]);;

(* polynme caractristique *)
let poca_2 m =
  let n = vect_length(m)       in
  let x = make_vect (n+1) zero in
  let y = make_vect (n+1) zero in
  for i=0 to n do
    x.(i) <- num_of_int i;
    y.(i) <- det(add_diag m (minus_num x.(i)))
  done;
  lagrange x y
;;

poca_2(m);;
